"""Link Integrity - link retriever methods.
"""
from plone.app.linkintegrity.interfaces import IRetriever
from plone.app.linkintegrity.parser import extractLinks
from plone.app.textfield import RichText
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import getAdditionalSchemata
from zope.component import adapter
from zope.component import getUtility
from zope.interface import implementer
from zope.schema import getFieldsInOrder


@implementer(IRetriever)
@adapter(IDexterityContent)
class DXGeneral:
    """General retriever for DX that extracts URLs from (rich) text fields."""

    def __init__(self, context):
        self.context = context

    def retrieveLinks(self):
        """Finds all links from the object and return them."""
        fti = getUtility(IDexterityFTI, name=self.context.portal_type)
        schema = fti.lookupSchema()
        additional_schema = getAdditionalSchemata(
            context=self.context, portal_type=self.context.portal_type
        )
        schemas = [i for i in additional_schema] + [schema]
        links = set()
        for schema in schemas:
            for name, field in getFieldsInOrder(schema):
                if isinstance(field, RichText):
                    value = getattr(schema(self.context), name)
                    if not value or not getattr(value, "raw", None):
                        continue
                    links |= set(extractLinks(value.raw))
        return links



from Acquisition import aq_inner
from datetime import datetime
from datetime import timedelta
from plone.app.linkintegrity.handlers import modifiedContent
from plone.base import PloneMessageFactory as _
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from transaction import savepoint
from zExceptions import NotFound

import logging


logger = logging.getLogger(__name__)


class UpdateView(BrowserView):
    """Iterate over all catalogued items and update linkintegrity-information."""

    def __call__(self):
        context = aq_inner(self.context)
        request = aq_inner(self.request)
        if "update" in request.form or "delete_all" in request.form:
            starttime = datetime.now()
            count = self.update()
            duration = timedelta(seconds=(datetime.now() - starttime).seconds)
            msg = _(
                "linkintegrity_update_info",
                default="Link integrity information updated for ${count} "
                + "items in ${time} seconds.",
                mapping={"count": count, "time": str(duration)},
            )
            IStatusMessage(request).add(msg, type="info")
            msg = "Updated {} items in {} seconds".format(
                count,
                str(duration),
            )
            logger.info(msg)
            request.RESPONSE.redirect(getToolByName(context, "portal_url")())
        elif "cancel" in request.form:
            msg = _("Update cancelled.")
            IStatusMessage(request).add(msg, type="info")
            request.RESPONSE.redirect(getToolByName(context, "portal_url")())
        else:
            return self.index()

    def update(self):
        catalog = getToolByName(self.context, "portal_catalog")
        count = 0

        for brain in catalog():
            try:
                obj = brain.getObject()
            except (AttributeError, NotFound, KeyError):
                msg = "Catalog inconsistency: {0} not found!"
                logger.error(msg.format(brain.getPath()), exc_info=1)
                continue
            try:
                modifiedContent(obj, "dummy event parameter")
                count += 1
            except Exception:
                msg = "Error updating linkintegrity-info for {0}."
                logger.error(msg.format(obj.absolute_url()), exc_info=1)
            if count % 1000 == 0:
                savepoint(optimistic=True)
        return count



# tests package
# to run all tests type '.../bin/instance test -s plone.app.linkintegrity'

# disable deprecation warnings
import warnings


warnings.filterwarnings("ignore", ".*", DeprecationWarning)



from plone.app.linkintegrity import testing
from plone.app.linkintegrity.browser.info import DeleteConfirmationInfo
from plone.app.linkintegrity.testing import create
from plone.app.linkintegrity.tests.utils import set_text
from plone.app.linkintegrity.utils import getOutgoingLinks
from plone.app.linkintegrity.utils import hasIncomingLinks
from plone.app.textfield import RichTextValue
from zope.lifecycleevent import modified

import unittest


class CircularReferencesTestCase(unittest.TestCase):
    """Circular reference testcase"""

    layer = testing.PLONE_APP_LINKINTEGRITY_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def _set_text(self, obj, text):
        obj.text = RichTextValue(text)
        modified(obj)

    def test_circular_reference_manages_relations(self):
        doc1 = self.portal["doc1"]
        doc2 = self.portal["doc2"]
        doc3 = self.portal["doc3"]
        self.assertFalse(hasIncomingLinks(doc1))
        self.assertFalse(hasIncomingLinks(doc2))
        self.assertFalse(hasIncomingLinks(doc3))
        set_text(doc1, '<a href="doc2">doc2</a>')
        set_text(doc2, '<a href="doc3">doc3</a>')
        set_text(doc3, '<a href="doc1">doc1</a>')
        self.assertTrue(hasIncomingLinks(doc1))
        self.assertTrue(hasIncomingLinks(doc2))
        self.assertTrue(hasIncomingLinks(doc3))

    def test_circular_reference_subfolder_deletion(self):
        doc1 = self.portal.doc1
        doc2 = self.portal.doc2
        doc3 = self.portal.doc3
        doc4 = self.portal.folder1.doc4
        folder1 = self.portal.folder1

        # This tests the behaviour when removing objects
        # referencing each other in a circle.
        set_text(doc1, '<a href="doc2">documents...</a>')
        set_text(doc2, '<a href="doc3">go round...</a>')
        set_text(doc3, '<a href="folder1/doc4">and round.</a>')
        set_text(doc4, '<a href="../doc1">in circles.</a>')

        self.assertEqual([r.to_object for r in getOutgoingLinks(doc1)], [doc2])
        self.assertEqual([r.to_object for r in getOutgoingLinks(doc2)], [doc3])
        self.assertEqual([r.to_object for r in getOutgoingLinks(doc3)], [doc4])
        self.assertEqual([r.to_object for r in getOutgoingLinks(doc4)], [doc1])

        view = DeleteConfirmationInfo(self.portal, self.request)
        self.assertEqual(len(view.get_breaches([folder1])), 1)
        self.assertEqual(len(view.get_breaches([doc1, doc2, doc3, folder1])), 0)
        self.assertEqual(len(view.get_breaches([doc2, folder1])), 2)

    def test_internal_breaches_are_dropped(self):
        folder1 = self.portal.folder1
        create(folder1, "Document", id="doc5", title="Test Page 5")
        doc1 = self.portal.doc1
        doc4 = self.portal.folder1.doc4
        doc5 = self.portal.folder1.doc5
        set_text(doc1, '<a href="folder1">f1</a>')
        set_text(doc4, '<a href="doc5">d5</a><a href="../doc1">d1</a>')
        set_text(doc5, '<a href="../folder1">f1</a>')

        doc4_breaches = {r.to_object for r in getOutgoingLinks(doc4)}
        # the order of breaches is non-deterministic
        self.assertEqual({doc1, doc5}, doc4_breaches)
        self.assertEqual([r.to_object for r in getOutgoingLinks(doc5)], [folder1])
        self.assertEqual([r.to_object for r in getOutgoingLinks(doc1)], [folder1])
        view = DeleteConfirmationInfo(self.portal, self.request)
        self.assertEqual(len(view.get_breaches([doc4])), 0)
        self.assertEqual(len(view.get_breaches([doc5])), 1)
        self.assertEqual(len(view.get_breaches([doc4, doc5])), 0)
        self.assertEqual(len(view.get_breaches([folder1])), 1)
        self.assertEqual(len(view.get_breaches([doc1])), 1)
        self.assertEqual(len(view.get_breaches([doc1, folder1])), 0)

        view = folder1.restrictedTraverse("delete_confirmation")
        self.assertIn("Potential link breakage", view())
        view = folder1.restrictedTraverse("delete_confirmation_info")
        self.assertIn("Potential link breakage", view())
        view = doc4.restrictedTraverse("delete_confirmation")
        self.assertNotIn("Potential link breakage", view())
        view = doc4.restrictedTraverse("delete_confirmation_info")
        self.assertNotIn("Potential link breakage", view())



from Acquisition import aq_get
from Acquisition import aq_parent
from plone.app.linkintegrity.interfaces import IRetriever
from plone.app.linkintegrity.utils import ensure_intid
from plone.app.linkintegrity.utils import referencedRelationship
from plone.app.uuid.utils import uuidToCatalogBrain
from plone.base.interfaces import IEditingSchema
from plone.base.interfaces import IPloneSiteRoot
from plone.dexterity.interfaces import IDexterityContent
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from urllib.parse import unquote
from urllib.parse import urlsplit
from z3c.relationfield import RelationValue
from z3c.relationfield.event import _setRelation
from zc.relation.interfaces import ICatalog
from zExceptions import NotFound
from ZODB.POSException import ConflictError
from zope.component import getUtility
from zope.interface.interfaces import ComponentLookupError
from zope.intid.interfaces import IIntIds
from zope.publisher.interfaces import NotFound as ztkNotFound

import logging


logger = logging.getLogger(__name__)


def findObject(base, path):
    """traverse to given path and find the upmost object"""
    if path.startswith("/"):
        # Make an absolute path relative to the portal root
        obj = getToolByName(base, "portal_url").getPortalObject()
        portal_path = obj.absolute_url_path() + "/"
        if path.startswith(portal_path):
            path = path[len(portal_path) :]
    else:
        obj = aq_parent(base)  # relative urls start at the parent...

    components = path.split("/")

    # Support resolveuid/UID paths explicitly, without relying
    # on a view or skinscript to do this for us.
    if "resolveuid" in components:
        uid = components[components.index("resolveuid") + 1]
        # This may be a link to a page that once was published but not anymore,
        # or the current editor does not have View permission.
        # In that case uuidToObject(uid) could fail with Unauthorized.
        brain = uuidToCatalogBrain(uid)
        if brain is not None:
            # Note: _unrestrictedGetObject starts with an underscore,
            # but it is documented in ZCatalog.interfaces,
            # so should be safe to rely on.
            obj = brain._unrestrictedGetObject()
            if obj:
                return obj, path

    while components:
        child_id = unquote(components[0])
        try:
            try:
                child = obj.unrestrictedTraverse(child_id)
            except AttributeError:
                request = aq_get(obj, "REQUEST")
                child = request.traverseName(obj, child_id)
        except ConflictError:
            raise
        except (AttributeError, KeyError, NotFound, ztkNotFound, UnicodeEncodeError):
            return None, None
        if not IDexterityContent.providedBy(child) and not IPloneSiteRoot.providedBy(
            child
        ):
            break
        obj = child
        components.pop(0)
    return obj, "/".join(components)


def getObjectsFromLinks(base, links):
    """Determine actual objects referred to by given links.

    return set of RelationValue
    """
    intids = getUtility(IIntIds)
    objects = set()
    url = base.absolute_url()
    scheme, host, path, query, frag = urlsplit(url)
    for link in links:
        s, h, path, q, f = urlsplit(link)
        # relative or local url
        if (not s and not h) or (s == scheme and h == host):
            obj, extra = findObject(base, path)
            if obj and not IPloneSiteRoot.providedBy(obj):
                objid = ensure_intid(obj, intids)
                if objid is None:
                    # if we get a NotYet error, the object is not
                    # attached yet and we will need to get links
                    # at a later time when the object has an intid
                    continue
                relation = RelationValue(objid)
                objects.add(relation)
    return objects


def modifiedContent(obj, event):
    """Object was modified, cloned or created."""
    if not check_linkintegrity_dependencies(obj):
        return
    retriever = IRetriever(obj, None)
    if retriever is not None:
        links = retriever.retrieveLinks()
        refs = getObjectsFromLinks(obj, links)
        updateReferences(obj, refs)


def removedContent(obj, event):
    if not check_linkintegrity_dependencies(obj):
        return

    intids = getUtility(IIntIds)
    try:
        int_id = intids.getId(obj)
    except KeyError:
        return

    catalog = getUtility(ICatalog)
    rels = catalog.findRelations(
        {"from_id": int_id, "from_attribute": referencedRelationship}
    )
    for rel in list(rels):
        catalog.unindex(rel)


# BBB
modifiedArchetype = modifiedContent
modifiedDexterity = modifiedContent


def updateReferences(obj, refs):
    """Renew all linkintegritry-references.

    Search the zc.relation catalog for linkintegritry-references for this obj.
    Drop them all and set the new ones.
    TODO: Might be improved by not changing anything if the links are the same.
    """
    int_id = ensure_intid(obj)
    if int_id is None:
        return
    catalog = getUtility(ICatalog)
    # unpack the rels before deleting
    old_rels = [
        i
        for i in catalog.findRelations(
            {"from_id": int_id, "from_attribute": referencedRelationship}
        )
    ]
    for old_rel in old_rels:
        catalog.unindex(old_rel)
    for ref in refs:
        _setRelation(obj, referencedRelationship, ref)


def check_linkintegrity_dependencies(obj):
    try:
        reg = getUtility(IRegistry)
        editing_settings = reg.forInterface(IEditingSchema, prefix="plone")
    except (ComponentLookupError, KeyError):
        return False
    if not editing_settings.enable_link_integrity_checks:
        return False
    if not getToolByName(obj, "portal_url", None):
        # `getObjectFromLinks` is not possible without access
        # to `portal_url`
        return False
    try:
        getUtility(IIntIds)
        getUtility(ICatalog)
    except ComponentLookupError:
        # Linkintegrity not possible without zope.intid-
        # and zc.relation-catalog
        return False
    return True



from OFS.ObjectManager import BeforeDeleteException
from zope.interface import implementer
from zope.interface import Interface


class ILinkIntegrityNotificationException(Interface):
    """An exception indicating a prevented link integrity breach."""


@implementer(ILinkIntegrityNotificationException)
class LinkIntegrityNotificationException(BeforeDeleteException):
    """An exception indicating a prevented link integrity breach."""

    def __str__(self):
        args = self.args
        if args and isinstance(args, tuple):
            return repr(args[0])
        return super().__str__()



from plone.base.interfaces import IEditingSchema
from plone.registry.interfaces import IRegistry
from zc.relation.interfaces import ICatalog
from zope.component import getUtility
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import NotYet


referencedRelationship = "isReferencing"


def getIncomingLinks(obj=None, intid=None, from_attribute=referencedRelationship):
    """Return a generator of incoming relations created using
    plone.app.linkintegrity (Links in Richtext-Fields).
    """
    catalog = getUtility(ICatalog)
    intid = intid if intid is not None else getUtility(IIntIds).getId(obj)
    query = {"to_id": intid}
    if from_attribute:
        query["from_attribute"] = from_attribute
    return catalog.findRelations(query)


def hasIncomingLinks(obj=None, intid=None):
    """Test if an object is linked to by other objects using
    plone.app.linkintegrity (Links in Richtext-Fields).

    Way to give bool without loading generator into list.
    """
    for it in getIncomingLinks(obj=obj, intid=intid):
        return True
    return False


def getOutgoingLinks(obj=None, intid=None, from_attribute=referencedRelationship):
    """Return a generator of outgoing relations created using
    plone.app.linkintegrity (Links in Richtext-Fields).
    """
    catalog = getUtility(ICatalog)
    intid = intid if intid is not None else getUtility(IIntIds).getId(obj)
    query = {"from_id": intid}
    if from_attribute:
        query["from_attribute"] = from_attribute
    return catalog.findRelations(query)


def hasOutgoingLinks(obj=None, intid=None):
    """Test if an object links to other objects using plone.app.linkintegrity
    (Links in Richtext-Fields).
    """
    for it in getOutgoingLinks(obj=obj, intid=intid):
        return True
    return False


def linkintegrity_enabled():
    reg = getUtility(IRegistry)
    editing_settings = reg.forInterface(IEditingSchema, prefix="plone")
    return editing_settings.enable_link_integrity_checks


def ensure_intid(obj, intids=None):
    if intids is None:
        intids = getUtility(IIntIds)
    try:
        obj_id = intids.getId(obj)
    except KeyError:
        # In some cases a object is not yet registered by the intid catalog
        try:
            obj_id = intids.register(obj)
        except NotYet:
            # if we get a NotYet error, the object is not
            # attached yet and we will need to get links
            # at a later time when the object has an intid
            return
    return obj_id



from plone.app.linkintegrity import testing
from plone.app.linkintegrity.handlers import findObject
from plone.app.linkintegrity.testing import create
from plone.app.testing import logout

import unittest


class ReferenceGenerationTestCase(unittest.TestCase):
    """testing the handlers.findObject function"""

    layer = testing.PLONE_APP_LINKINTEGRITY_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def test_relative_to_portal_root_1(self):
        obj, components = findObject(self.portal.doc1, "/plone/doc2")
        self.assertEqual(obj.absolute_url_path(), "/plone/doc2")
        self.assertEqual(components, "")

    def test_relative_to_portal_root_2(self):
        # Prevent regression.
        # See https://github.com/plone/plone.app.linkintegrity/pull/17
        obj, components = findObject(self.portal.doc1, "/doc2")
        self.assertEqual(obj.absolute_url_path(), "/plone/doc2")
        self.assertEqual(components, "")

    def test_webserver_rewrites_portal_name(self):
        # test the case where a webserver rewrites the portal name,
        # e.g. for Apache:
        # RewriteRule ^/wssitename(.*)$ http://localhost:8080/VirtualHostBase/http/my.domain.com:80/plonesitename/VirtualHostRoot/_vh_wssitename$1  # noqa
        self.portal.REQUEST.other["VirtualRootPhysicalPath"] = ("", "plone")
        self.portal.REQUEST._script = ["plone_foo"]
        obj, components = findObject(self.portal.doc1, "/plone_foo/doc2")
        self.assertEqual(obj.absolute_url_path(), "/plone_foo/doc2")
        self.assertEqual(obj.getPhysicalPath(), ("", "plone", "doc2"))
        self.assertEqual(components, "")

    def test_uuid_link(self):
        # Test that objects can be found from uuid links.
        create(self.portal, "Document", id="target", title="Target")
        target = self.portal.target
        target_uid = target.UID()
        path = f"../resolveuid/{target_uid}"

        # We logout.  This is to check that findObject also finds objects
        # that are not visible to the current user, like a private page.
        # See https://github.com/plone/plone.app.linkintegrity/issues/79
        logout()

        obj, components = findObject(self.portal.doc1, path)
        self.assertEqual(obj.absolute_url_path(), "/plone/target")
        self.assertEqual(components, path)



from Acquisition import aq_inner
from OFS.interfaces import IFolder
from plone.app.linkintegrity.utils import getIncomingLinks
from plone.app.linkintegrity.utils import linkintegrity_enabled
from plone.base import PloneMessageFactory as _
from plone.base.interfaces import IPloneSiteRoot
from plone.uuid.interfaces import IUUID
from Products.CMFCore.permissions import AccessContentsInformation
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.i18n import translate

import logging


logger = logging.getLogger(__name__)


class DeleteConfirmationInfo(BrowserView):
    template = ViewPageTemplateFile("delete_confirmation_info.pt")
    breach_count = {}

    def __init__(self, context, request):
        self.linkintegrity_enabled = linkintegrity_enabled()
        self.context = context
        self.request = request

    def __call__(self, items=None):
        if not self.linkintegrity_enabled:
            return
        if items is None:
            if IPloneSiteRoot.providedBy(self.context):
                # Checking the portal for breaches makes no sense.
                return
            else:
                items = [self.context]
        self.breaches = self.get_breaches(items)
        return self.template()

    def get_breaches(self, items=None):
        """Return breaches for multiple items.

        Breaches coming from objects in the list of items
        or their children (if a object is a folder) will be ignored.
        """
        if items is None:
            items = [self.context]
        catalog = getToolByName(self.context, "portal_catalog")
        results = []
        uids_to_ignore = []
        uids_visited = set()
        self.breach_count = {}
        for obj in items:
            obj_path = "/".join(obj.getPhysicalPath())
            brains_to_delete = catalog(path={"query": obj_path})
            # add the current items uid and all its children's uids to the
            # list of uids that are ignored
            uids_to_ignore.extend([i.UID for i in brains_to_delete])
            for brain_to_delete in brains_to_delete:
                try:
                    obj_to_delete = brain_to_delete.getObject()  # noqa
                except (AttributeError, KeyError):
                    logger.exception(
                        "No object found for %s! Skipping", brain_to_delete
                    )
                    continue
                for breach in self.get_breaches_for_item(obj):
                    add_breach = False
                    for source in breach["sources"]:
                        # Only add the breach if one the sources is not in the
                        # list of items that are to be deleted.
                        if (
                            source["uid"] not in uids_to_ignore
                            and source["uid"] not in uids_visited
                        ):
                            add_breach = True
                            uids_visited.add(source["uid"])
                            break
                    if add_breach:
                        results.append(breach)
            if IFolder.providedBy(obj):
                count = len(catalog(path={"query": obj_path}))
                count_dirs = len(catalog(path={"query": obj_path}, is_folderish=True))
                count_public = len(
                    catalog(path={"query": obj_path}, review_state="published")
                )
                if count:
                    self.breach_count[obj_path] = [count, count_dirs, count_public]

        # Cleanup: Some breaches where added before it was known
        # that their source will be deleted too.
        for result in results:
            for source in result["sources"]:
                if source["uid"] in uids_to_ignore:
                    # Drop sources that are also being deleted
                    result["sources"].remove(source)
                    if not result["sources"]:
                        # Remove the breach is there are no more sources
                        # This check is necessary since there can be multiple
                        # sources for a breach
                        results.remove(result)
        return results

    def get_breaches_for_item(self, obj=None):
        """Get breaches for one object and its children.

        Breaches coming from the children of a folder are ignored by default.
        """
        if obj is None:
            obj = self.context
        results = []
        catalog = getToolByName(obj, "portal_catalog")
        obj_path = "/".join(obj.getPhysicalPath())

        breaches = self.check_object(obj)
        if breaches:
            results.append(breaches)

        if IFolder.providedBy(obj):
            brains = catalog(path={"query": obj_path})
            for brain in brains:
                try:
                    child = brain.getObject()
                except (AttributeError, KeyError):
                    continue
                if child == obj:
                    continue
                breaches = self.check_object(obj=child, excluded_path=obj_path)
                if breaches:
                    results.append(breaches)
        self.breaches = results
        return results

    def check_object(self, obj, excluded_path=None):
        """Check one object for breaches.
        Breaches originating from excluded_path are ignored.
        """
        breaches = {}
        direct_links = getIncomingLinks(obj, from_attribute=None)
        has_breaches = False
        for direct_link in direct_links:
            source_path = direct_link.from_path
            if not source_path:
                # link is broken
                continue
            if excluded_path and source_path.startswith(excluded_path):
                # source is in excluded_path
                continue
            source = direct_link.from_object
            if not breaches.get("sources"):
                breaches["sources"] = []
            breaches["sources"].append(
                {
                    "uid": IUUID(source),
                    "title": source.Title(),
                    "url": source.absolute_url(),
                    "accessible": self.is_accessible(source),
                }
            )
            has_breaches = True
        if has_breaches:
            breaches["target"] = {
                "uid": IUUID(obj),
                "title": obj.Title(),
                "url": obj.absolute_url(),
                "portal_type": obj.portal_type,
                "type_title": self.get_portal_type_title(obj),
            }
            return breaches

    def get_portal_type_title(self, obj):
        """Get the portal type title of the object."""
        context = aq_inner(self.context)
        portal_types = getToolByName(context, "portal_types")
        fti = portal_types.get(obj.portal_type)
        if fti is not None:
            type_title_msgid = fti.Title()
        else:
            type_title_msgid = obj.portal_type
        type_title = translate(type_title_msgid, context=self.request)
        return type_title

    def is_accessible(self, obj):
        return _checkPermission(AccessContentsInformation, obj)

    def objects(self):
        return [_("Objects in all"), _("Folders"), _("Published objects")]



from base64 import decodebytes
from plone.app.testing import layers
from plone.app.testing import login
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.namedfile.file import NamedImage
from Products.CMFCore.utils import getToolByName
from zope.configuration import xmlconfig

import io


B64_DATA = b"R0lGODlhAQABAIAAAAUEBAAAACwAAAAAAQABAAACAkQBADs="
GIF = io.BytesIO(decodebytes(B64_DATA))
GIF.filename = "sample.gif"
GIF.contentType = "image/gif"
GIF._width = 1
GIF._height = 1


def create(container, type_name, **kwargs):
    """A easy helper method to create some content since we do not have
    plone.api in core.
    """
    new_id = container.invokeFactory(type_name, **kwargs)
    content = container[new_id]
    return content


class LinkIntegrityLayer(PloneSandboxLayer):
    """Base Layer for Dexterity testing."""

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import plone.app.linkintegrity

        xmlconfig.file(
            "configure.zcml", plone.app.linkintegrity, context=configurationContext
        )

    def setUpPloneSite(self, portal):
        setRoles(
            portal,
            TEST_USER_ID,
            [
                "Manager",
            ],
        )
        login(portal, TEST_USER_NAME)

        # Create sample documents
        type_data = dict(type_name="Document")
        for i in range(1, 4):
            type_data["id"] = f"doc{i:d}"
            type_data["title"] = f"Test Page {i:d}"
            create(portal, **type_data)

        create(portal, "File", id="file1", title="File 1", file=GIF)
        create(portal, "Folder", id="folder1", title="Folder 1")
        subfolder = portal["folder1"]
        create(subfolder, "Document", id="doc4", title="Test Page 4")

        # setup members
        pm = getToolByName(portal, "portal_membership")
        pm.addMember("editor", TEST_USER_PASSWORD, ["Editor"], [])
        pm.addMember("member", TEST_USER_PASSWORD, ["Member"], [])
        pm.addMember("authenticated", TEST_USER_PASSWORD, [], [])

        # Create an object that does not provide the behavior to live along
        create(portal, "News Item", id="news1", title="News 1")

        # create a NamedImage
        portal.invokeFactory("Image", "image1")
        portal["image1"].image = NamedImage(GIF, "image/gif", "sample.gif")


PLONE_APP_LINKINTEGRITY_FIXTURE = LinkIntegrityLayer()

PLONE_APP_LINKINTEGRITY_INTEGRATION_TESTING = layers.IntegrationTesting(
    bases=(PLONE_APP_LINKINTEGRITY_FIXTURE,), name="plone.app.linkintegrity:Integration"
)

PLONE_APP_LINKINTEGRITY_FUNCTIONAL_TESTING = layers.FunctionalTesting(
    bases=(PLONE_APP_LINKINTEGRITY_FIXTURE,), name="plone.app.linkintegrity:Functional"
)



from plone.app.linkintegrity import testing
from plone.app.linkintegrity.tests.utils import set_text
from plone.app.linkintegrity.utils import getIncomingLinks
from plone.app.linkintegrity.utils import getOutgoingLinks
from plone.uuid.interfaces import IUUID

import unittest


class ImageReferenceTestCase(unittest.TestCase):
    """image reference testcase"""

    layer = testing.PLONE_APP_LINKINTEGRITY_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def test_image_tag_reference_creation(self):
        doc1 = self.portal.doc1
        img1 = self.portal.image1

        # This tests the correct creation of references used for
        # ensuring link integrity. Any archetype-based content object
        # which refers to other (local) objects by `<img>` or `<a>` tags
        # should create references between those objects on save.
        set_text(doc1, img1.restrictedTraverse("@@images").tag())

        self.assertEqual(
            [r.to_object for r in getOutgoingLinks(doc1)],
            [
                img1,
            ],
        )
        self.assertEqual([r.to_object for r in getIncomingLinks(doc1)], [])
        self.assertEqual([r.to_object for r in getOutgoingLinks(img1)], [])
        self.assertEqual(
            [r.from_object for r in getIncomingLinks(img1)],
            [doc1],
        )

    def test_image_scale_reference_creation(self):
        doc1 = self.portal.doc1
        img1 = self.portal.image1

        # Linking image scales should also work:
        set_text(doc1, '<a href="image1/@@images/image_thumb">an image</a>')
        self.assertEqual(
            [r.to_object for r in getOutgoingLinks(doc1)],
            [
                img1,
            ],
        )
        self.assertEqual(
            [r.from_object for r in getIncomingLinks(img1)],
            [
                doc1,
            ],
        )

    def test_image_resolveuid_reference_creation(self):
        doc1 = self.portal.doc1
        img1 = self.portal.image1

        # Linking via the "resolveuid/UID" method should also work:
        set_text(doc1, f'<a href="resolveuid/{IUUID(img1):s}">an image</a>')
        self.assertEqual(
            [r.to_object for r in getOutgoingLinks(doc1)],
            [
                img1,
            ],
        )
        self.assertEqual(
            [r.from_object for r in getIncomingLinks(img1)],
            [
                doc1,
            ],
        )



from html.parser import HTMLParser


TAG_ATTRS_TO_TRACK = {
    # The humble hyperlink.
    "a": ["href"],
    # The image.
    "img": ["src", "srcset"],
    # Used within img/picture/audio/video tags
    # to embed various sources of media.
    "source": ["src", "srcset"],
    # Embeds audio recordings.
    "audio": ["src"],
    # Embeds videos.
    "video": ["src"],
    # Used to embed PDFs.
    "iframe": ["src"],
}


class LinkParser(HTMLParser):
    """A simple html parser for link and image urls."""

    def __init__(self):
        HTMLParser.__init__(self)
        self.links = []

    def getLinks(self):
        """Return all links found during parsing."""
        return tuple(self.links)

    def handle_starttag(self, tag, attrs):
        """Override the method to remember all links."""
        for at in TAG_ATTRS_TO_TRACK.get(tag.lower(), []):
            self.links.extend(search_attr(at, attrs))


def links_in_srcset(attrval):
    # SRCSET is split by commas, and each line's first
    # element is the URL in question.
    # Yes, this means that spaces in such a link must be
    # encoded with %20 or +.  That is what the written
    # standard implies.
    return [src.strip().split()[0] for src in attrval.split(",")]


def search_attr(name, attrs):
    """Search named attribute in a list of attributes."""
    for attr, value in attrs:
        if attr == name:
            return links_in_srcset(value) if name == "srcset" else [value]
    return []


def extractLinks(data, encoding="utf-8"):
    """Parse the given html and return all links."""
    if not data:
        return []
    parser = LinkParser()
    try:
        parser.feed(data)
        parser.close()
    except UnicodeDecodeError:
        parser = LinkParser()
        parser.feed(data.decode(encoding))
        parser.close()
    except TypeError:
        pass

    return parser.getLinks()
